package org.neo4j.community.console;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.TimeUnit;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.transaction.InvalidTransactionException;
import javax.transaction.SystemException;
import javax.transaction.TransactionManager;

import Translator.Q1;
import Translator.TrapezoidalFunction;

import org.neo4j.cypher.javacompat.ExecutionResult;
import org.neo4j.cypher.javacompat.PlanDescription;
import org.neo4j.cypher.javacompat.QueryStatistics;
import org.neo4j.cypher.javacompat.internal.ServerExecutionEngine;
import org.neo4j.graphdb.GraphDatabaseService;
import org.neo4j.graphdb.Node;
import org.neo4j.graphdb.PropertyContainer;
import org.neo4j.graphdb.Relationship;
import org.neo4j.helpers.collection.IteratorUtil;
import org.neo4j.helpers.collection.MapUtil;
import org.neo4j.kernel.GraphDatabaseAPI;
import org.neo4j.kernel.impl.util.StringLogger;

import scala.NotImplementedError;

/**
 * @author mh
 * @since 21.04.12
 */
public class CypherQueryExecutor {
	private static final Pattern PROPERTY_PATTERN = Pattern.compile("((\\w+)\\s*:|\\w+\\.(\\w+)\\s*=)",Pattern.MULTILINE|Pattern.DOTALL);
	private static final Pattern INDEX_PATTERN = Pattern.compile("(node|relationship)\\s*:\\s*(\\p{javaJavaIdentifierStart}\\p{javaJavaIdentifierPart}+|`[^`]+`|)\\s*\\(",Pattern.MULTILINE);
	public static final Pattern CANNOT_PROFILE_PATTERN = Pattern.compile("\\b(UNION|OPTIONAL|LOAD)\\b|(\\bMERGE\\b.+){2,}", Pattern.CASE_INSENSITIVE|Pattern.MULTILINE|Pattern.DOTALL);
	private final TransactionManager transactionManager;
	private ServerExecutionEngine executionEngine;
	private final Index index;
	private final GraphDatabaseService gdb;
	public static final int CYPHER_LENGTH = "CYPHER".length();

	public CypherQueryExecutor(GraphDatabaseService gdb, Index index) {
		this.gdb = gdb;
		transactionManager = ((GraphDatabaseAPI) gdb).getDependencyResolver().resolveDependency(TransactionManager.class);
		this.index = index;
		executionEngine = new ServerExecutionEngine(gdb, StringLogger.SYSTEM);
	}

	public boolean isMutatingQuery(String query) {
		return query.matches("(?is).*\\b(create|relate|merge|delete|set)\\b.*");
	}
	public boolean isIndexQuery(String query) {
		return query.matches("(?is).*\\bcreate (index|constraint)\\b.*");
	}
	public boolean isCypherQuery(String query) {
		return query.matches("(?is).*\\b(drop|start|merge|match|return|where|skip|limit|create|relate|delete|set)\\b.*");
	}

	public static class CypherResult implements Iterable<Map<String, Object>> {
		private List<String> columns;
		private final String query;
		private String text;
		//private String atsData;
		private List<Map<String, Object>> rows;
		private List<Map<String,Object>> mutable_rows;
		private List<Map<String, Object>> json;
		private QueryStatistics queryStatistics;
		private final PlanDescription plan;
		private final long time;
		static final String SAT_DEGREE_COL_NAME = "Satisfaction_degree";
	
		public CypherResult(List<String> columns, List<Map<String, Object>> rows, QueryStatistics queryStatistics, long time, PlanDescription plan, String query) {
			this.query = query; 
			this.columns = new ArrayList<>(columns);
			this.queryStatistics = queryStatistics;
			this.time = time;
			this.rows = rows;
			this.plan = plan;
			this.text = generateText(); 
			this.json = createJson();
			this.mutable_rows = new LinkedList<Map<String,Object>>();

			// As row is immutable, we create a copy of row in a mutable object mutable_rows of type LinkedList<HashMap<String,Object>>
			Iterator it=rows.iterator();
			while (it.hasNext()) // For each row
			{
				Object the_current_row = it.next();
			//	System.out.println("the_current_row "+ the_current_row);
				HashMap<String,Object> new_mutable_row = new HashMap<String,Object>();
				
				Set listKeys=((Map<String, Object>) the_current_row).keySet();  // Get the list of keys in the map
				Iterator iterateur=listKeys.iterator();
				// Print the entry for each key
				while(iterateur.hasNext())
				{
					Object key= iterateur.next();
					new_mutable_row.put((String) key, ((Map<String, Object>) the_current_row).get(key));
				}
				mutable_rows.add(new_mutable_row);
			}
			//println_rows();
			//println_mutable_rows();
		}

		public void println_mutable_rows(){
			System.out.println("begin println_mutable_rows()");
			Iterator it=mutable_rows.iterator();
			while (it.hasNext()) // For each row
			{
				Object the_current_row = it.next();
				System.out.println ("Answer row:");
				// Content of the row (a row = a map) 
				Set listKeys=((Map<String, Object>) the_current_row).keySet();  // Get the list of keys in the map
				Iterator iterateur=listKeys.iterator();
				// Print the entry for each key
				while(iterateur.hasNext())
				{
					Object key= iterateur.next();
					System.out.println ("[Mutable rows] "+key+"=>"+((Map<String, Object>) the_current_row).get(key)+"(type "+ ((Map<String, Object>) the_current_row).get(key).getClass()+")");
				}

			}
			System.out.println("end println_mutable_rows()");
		}

		public void println_rows(){
			System.out.println("begin println_rows()");
			Iterator it=rows.iterator();
			while (it.hasNext()) // For each row
			{
				Object the_current_row = it.next();
				System.out.println ("Answer row:");
				// Content of the row (a row = a map) 
				Set listKeys=((Map<String, Object>) the_current_row).keySet();  // Get the list of keys in the map
				Iterator iterateur=listKeys.iterator();
				// Print the entry for each key
				while(iterateur.hasNext())
				{
					Object key= iterateur.next();
					System.out.println ("[Immutable rows] "+key+"=>"+((Map<String, Object>) the_current_row).get(key)+"(type "+ ((Map<String, Object>) the_current_row).get(key).getClass()+")");
				}

			}
			System.out.println("end println_rows()");
		}
		
		// ADDED FOR QUANTIFIED QUERY
		public void Quant_score_calculator_module(Map <String, Double> List){
			//System.out.println("begin fuzzy_sat_degree");
			Q1 q1= new Q1();
			String firstNode=q1.firstNode;
			List<Map<String,Object>> mutable_rows_copie = new LinkedList<Map<String,Object>>();
			Iterator it1=mutable_rows.iterator();
			ArrayList<String> ListNode = new ArrayList<String>();
			HashMap UniqueList =new HashMap<Object, Double>();
			 HashMap<String,Object> new_mutable_row = null;
			 Node node;
			 Object nameNode = null;
	            
			while (it1.hasNext()) { // For each row
				Object N_mutable_row = it1.next();
				//System.out.println("the_current_row row "+ N_mutable_row.toString());
				node = (Node) ((Map<String, Object>) N_mutable_row).get(firstNode);
				Double sat = (Double) ((Map<String, Object>) N_mutable_row).get("Satisfaction_degree");
		    	if (!UniqueList.containsKey(node)){
					    	  UniqueList.put(node, sat);
					    	  new_mutable_row = new HashMap<String,Object>();
		                      Set listKeys=((Map<String, Object>) N_mutable_row).keySet();  // Get the list of keys in the map
		                      Iterator iterateur=listKeys.iterator();
					        	// Print the entry for each key
					          while(iterateur.hasNext()){
					        		Object key= iterateur.next();
					        		new_mutable_row.put((String) key, ((Map<String, Object>) N_mutable_row).get(key));
					        		//	System.out.println("New_mutable_row "+ new_mutable_row.toString());
					        				}
					        		mutable_rows_copie.add(new_mutable_row);			
					      }}
			
			       Iterator it2=mutable_rows_copie.iterator();
					     while (it2.hasNext()) { // For each row
								Object row = it2.next();      
								node = (Node) ((Map<String, Object>) row).get(firstNode);		 
						    	nameNode= node.getProperty("name");
						    	Double FinalDegree= List.get(nameNode);    	
		    	         ((Map<String, Object>) row).put(SAT_DEGREE_COL_NAME, FinalDegree);
						}	 
				
			//3- Round the sat degrees at 2 decimals and
						Iterator<Map<String, Object>> it3=mutable_rows_copie.iterator();
						while (it3.hasNext()) {// For each row
							HashMap<String, Object> the_current_row = (HashMap<String, Object>) it3.next();
							Set<String> listKeys=(the_current_row).keySet();  // Get the list of columns names in the map
							Iterator<String> i4=listKeys.iterator(); 
							while(i4.hasNext()) {// For each column named calc_fuzzy_*
								String nom_col = i4.next();
								if (nom_col.equals(SAT_DEGREE_COL_NAME)) {
															
									String convertedString = String.format(Locale.US, "%.2f", ((Double) the_current_row.get(nom_col)));
									Double roundedValue = Double.valueOf(convertedString);
									the_current_row.put(SAT_DEGREE_COL_NAME, roundedValue);

								}		
							}		
						}

						// 4- Sorting according to the satisfaction degree
						// THIS PART OF THE CODE COULD BE OPTIMIZED BY CONSIDERING STORING THE LIST OF ANSWERS IN A LIST OF ROWS INSTEAD OF LIST OF  Map<String, Object>, IN ALLT THE CODE. THIS OPTIMIZATION HAS NOT BEEN IMPLEMENTED YET.
						// Sorting is composed of three stages : 4.1) Convert the mutable_rows (List of Map<String, Object> instances) into a List of Row instances. This is needed as the Row class implements Comparable interface and then, 4.2) Then the List of Row instances can be sorted. 4.3) At last the list of Rows instances is converted back into a list of Map<String, Object> instances.
						LinkedList<Row> list_of_Rows = new LinkedList<Row>();
						// 4.1
						Iterator<Map<String, Object>> it4=mutable_rows_copie.iterator();
						while (it4.hasNext()) {// For each element of mutable_rows
							Row new_row = new Row(it4.next());
							
							list_of_Rows.add(new_row);
						}
						// 4.2
						Collections.sort(list_of_Rows);
						Collections.reverse(list_of_Rows);
										// 4.3
						mutable_rows_copie.clear();
						Iterator<Row> it5=list_of_Rows.iterator();
						while (it5.hasNext()) { mutable_rows_copie.add(it5.next().getValue()); }
                        
						rows = mutable_rows_copie;
						this.text = generateText(); 
						this.json = createJson();
						//System.out.println("end fuzzy_sat_degree");
	}

		// ADDED FOR SUGAR
		public void fudge_score_calculator_module(Map <String,TrapezoidalFunction> trapezList){
		//	System.out.println("begin fuzzy_sat_degree");

			columns.add(SAT_DEGREE_COL_NAME);

			// 1- For each row, we add the default value 1 to the column nammed "satisfaction degree"
			Iterator<Map<String, Object>> it1=mutable_rows.iterator();
			while (it1.hasNext()) { // For each row
				Object mutable_row = it1.next();
				((HashMap<String, Object>) mutable_row).put(SAT_DEGREE_COL_NAME, (Double) 1.0);
				//System.out.println("[SUGAR DEBUG:]"+((Map<String, Object>) mutable_row).get(SAT_DEGREE_COL_NAME).getClass());
			}

			// 2- For each row, update the satisfaction degree
			Iterator<Map<String, Object>> it2=mutable_rows.iterator();
			while (it2.hasNext()) {// For each row
				HashMap<String, Object> the_current_row = (HashMap<String, Object>) it2.next();

				Set<String> listKeys=(the_current_row).keySet();  // Get the list of columns names in the map
				Iterator<String> i2=listKeys.iterator(); 
				while(i2.hasNext()) // For each column named fudge_calc_*
				{
					String nom_col = i2.next();
					if (nom_col.startsWith("fudge_calc_")) {
						//System.out.println("[nom_col_fuzzy]"+nom_col);
						String fuzzy_function = nom_col.substring(nom_col.lastIndexOf("_")+1);
						//System.out.println("[fuzzy_function]"+trapezList.get(fuzzy_function).toString());
						Double current_sat_degree = (Double) the_current_row.get(SAT_DEGREE_COL_NAME);
						//System.out.println("[current_sat_degree]"+current_sat_degree);
						//System.out.println("[the_current_row.get(nom_col)]"+the_current_row.get(nom_col)+" classe :"+(the_current_row.get(nom_col)).getClass());
						Double value_in_fuzzy_column;
						if (the_current_row.get(nom_col) instanceof Long)
						{value_in_fuzzy_column=((Long) the_current_row.get(nom_col)).doubleValue();}
						else{value_in_fuzzy_column= (Double) the_current_row.get(nom_col);};
						//System.out.println("[value_in_fuzzy_column]"+value_in_fuzzy_column);
						Double column_sat_degree = (trapezList.get(fuzzy_function)).mu(value_in_fuzzy_column); 
						//System.out.println("[column_sat_degree]"+column_sat_degree);
						the_current_row.put(SAT_DEGREE_COL_NAME, Math.min(current_sat_degree,column_sat_degree));
						i2.remove(); // Deletes the intermediate column named calc_fuzzy_*
						columns.remove(nom_col); // Deletes to columns name in columns
					}
				}
			}

			// 3- Round the sat degrees at 2 decimals and
			Iterator<Map<String, Object>> it3=mutable_rows.iterator();
			while (it3.hasNext()) {// For each row
				HashMap<String, Object> the_current_row = (HashMap<String, Object>) it3.next();
				Set<String> listKeys=(the_current_row).keySet();  // Get the list of columns names in the map
				Iterator<String> i4=listKeys.iterator(); 
				while(i4.hasNext()) {// For each column named calc_fuzzy_*
					String nom_col = i4.next();
					if (nom_col.equals(SAT_DEGREE_COL_NAME)) {
												
						String convertedString = String.format(Locale.US, "%.2f", ((Double) the_current_row.get(nom_col)));
						Double roundedValue = Double.valueOf(convertedString);
						the_current_row.put(SAT_DEGREE_COL_NAME, roundedValue);

					}		
				}		
			}

			// 4- Sorting according to the satisfaction degree
			// THIS PART OF THE CODE COULD BE OPTIMIZED BY CONSIDERING STORING THE LIST OF ANSWERS IN A LIST OF ROWS INSTEAD OF LIST OF  Map<String, Object>, IN ALLT THE CODE. THIS OPTIMIZATION HAS NOT BEEN IMPLEMENTED YET.
			// Sorting is composed of three stages : 4.1) Convert the mutable_rows (List of Map<String, Object> instances) into a List of Row instances. This is needed as the Row class implements Comparable interface and then, 4.2) Then the List of Row instances can be sorted. 4.3) At last the list of Rows instances is converted back into a list of Map<String, Object> instances.
			LinkedList<Row> list_of_Rows = new LinkedList<Row>();
			// 4.1
			Iterator<Map<String, Object>> it4=mutable_rows.iterator();
			while (it4.hasNext()) {// For each element of mutable_rows
				Row new_row = new Row(it4.next());
				list_of_Rows.add(new_row);
			}
			// 4.2
			Collections.sort(list_of_Rows);
			Collections.reverse(list_of_Rows);
			// 4.3
			mutable_rows.clear();
			Iterator<Row> it5=list_of_Rows.iterator();
			while (it5.hasNext()) { mutable_rows.add(it5.next().getValue()); }

			rows = mutable_rows;
			this.text = generateText(); 
			this.json = createJson();
			//System.out.println("end fuzzy_sat_degree");
		}
		
		/* deduplicate the list of rows : finalRows*/
		public List<Map <String,Object>> deduplicate(List<Map <String,Object>> finalRows1){

			//the list of rows without duplicates 
			List <Map <String,Object>> newList = new LinkedList<Map<String,Object>>();
		
			//iterate through the list of maps : finalRows
			for(int k=0;k<finalRows1.size();k++){
				// and put index variable in the buffer and compare it to other elements of the list of maps (finalRows)
				int index = finalRows1.indexOf(finalRows1.get(k)); 
				//iterate through the list of maps : finalRows  
				for(int l=0; l<finalRows1.size();l++){
					// ensure the difference between index and the new element token in the first list(finalRows)
					if(l !=index && ! finalRows1.get(index).equals(finalRows1.get(l))){ 
						//the new element token must not be in the list(newList) to avoid duplicates 
						if(isInList(finalRows1.get(l), newList)== false){ 
							System.out.println("Yesss it is"+finalRows1.get(l));	
							//add this famous element to the newList
							newList.add(finalRows1.get(l));	
						}
					}
				}
			}
			/*In the case when newList is empty ~ a unique element in finalRows
			 * we must add this element to newList
			 * because there is no other other element in 
			 * finalRows with different properties to compare with */
			
			if(!finalRows1.isEmpty() && newList.isEmpty()){
				newList.add(finalRows1.get(0));
			}
			return newList;
		}
		
		
		public List <Map <String,Object>> descendingSort(List <Map <String,Object>> mutable_rows1){	
			ArrayList<Double> maListe = degreeSortList(mutable_rows1); //Get the sorted list of satisfaction degrees (descending way)
			Map<String, Object> goLine = new HashMap<String , Object>(); // The searched row
			List <Map <String,Object>> finalRows = new LinkedList<Map<String,Object>>(); //The final result (finalRows)
			for(int i=0; i<maListe.size();i++){	//For each satisfaction_degree in the sorted "maListe"
				Double myDegree = maListe.get(i);
				goLine = getRowByDegree(myDegree,(LinkedList<Map<String,Object>>) mutable_rows1); // Get the row by its degree (myDegree)
				//before we add goLine to the final result, we should very if it is already in the finalResult
				if(isInList(goLine, finalRows)== false){
					finalRows.add(goLine); //Add the found row in in the final collection
				}
				else{
					System.out.println("the row "+goLine+" is already in finalRows	");
				}
			}
			return finalRows;
		}
			
			/* @param : Collection<Map<String, Object>> maCollection
		 * @return : ArrayList<Double> arrayOfDegrees
		 * Description : This function takes a collection of rows in the argument 
		 * 				 and retrieve The Satisfaction degree values from the column (key in hashMap) 
		 * 				 Satisfaction_degree and put those values in an ArraList called arrayOfDegrees */
		
		public ArrayList<Double> degreeSortList(Collection<Map<String, Object>> maCollection){
			/*Create an arrayList of satisfaction degrees gathered from the
			 * column Satisfaction degree*/
			ArrayList<Double> arrayOfDegrees = new ArrayList<Double>();  
			Iterator<Map<String, Object>> it = maCollection.iterator();
			while(it.hasNext()){ // For each row in maCollection
				
				Map<String, Object> maLigne = it.next(); // The current row
				Double satisfactionDegree = (Double) maLigne.get(SAT_DEGREE_COL_NAME); // Get the value in the column SAT_DEGREE_COL_NAME
				arrayOfDegrees.add(satisfactionDegree); // Add  "satisfactionDegree" to the arrayList of "satisfaction degrees" 
			}
			Comparator<Double> comparator = Collections.reverseOrder(); // Descending order
			Collections.sort(arrayOfDegrees,comparator); // Sort the arrayOfDegrees in the descending way
			return arrayOfDegrees;

		}
		
		public boolean isInList(Map<String, Object> row, List <Map<String, Object>>  maListe){
			boolean trouve =false;
			for(Map<String, Object> currentRow : maListe ){
				if(currentRow.equals(row)){
					trouve = true;
				}
				else{
					trouve = false;
				}
			}
			return trouve;
		}
		
		/* @param 		: 	- Double satisfaction : the satisfaction degree
		 * 					- LinkedList<Map<String, Object>> collection
		 * @return 		: 	Map<String, Object> foundRow : the row retrieved thanks to "satisfaction" value
		 * Description 	:	the function getRowByDegree enables to get the row we're looking for
		 * 					thanks to the value which is the parameter of this function   */
		
		public Map<String, Object> getRowByDegree(Double satisfaction, LinkedList<Map<String, Object>> collection ){
			Map<String, Object> foundRow = new HashMap<String, Object> (); // The searched row
			Map<String, Object> maLigne = new HashMap<String, Object> ();//The row of collection
			for(int i=0; i< collection.size(); i++){ // For each row in the collection
				maLigne = collection.get(i);
				if(maLigne.get(SAT_DEGREE_COL_NAME).equals(satisfaction)){ // Match of satisfaction degree
						foundRow = maLigne;
				}
			}
			return foundRow;
		}

		
		public List<String> getColumns() {
			return columns;
		}
		

		public String getQuery() {
			return query;
		}

		public int getRowCount() {
			return rows.size();
		}

		public Map getQueryStatistics() {
			final Map<String, Object> stats = MapUtil.map(
					"rows", getRowCount(),
					"time", getTime()
					);
			if (queryStatistics!=null && queryStatistics.containsUpdates()) {
				stats.put("containsUpdates", queryStatistics.containsUpdates());
				stats.put("nodesDeleted", queryStatistics.getDeletedNodes());
				stats.put("relationshipsDeleted", queryStatistics.getDeletedRelationships());
				stats.put("nodesCreated", queryStatistics.getNodesCreated());
				stats.put("relationshipsCreated", queryStatistics.getRelationshipsCreated());
				stats.put("propertiesSet", queryStatistics.getPropertiesSet());
				stats.put("text", queryStatistics.toString());
			}
			return stats;
		}


		public String getText() {
			return text;
		}

		private String generateText() {
			return new ResultPrinter().generateText(columns, rows, time, queryStatistics);
		}

		public Collection<Map<String, Object>> getRows() {
			return rows;
		}

		public String getPlan() {
			return plan!=null ? plan.toString() : "No Query Plan";
		}

		@Override
		public String toString() {
			return getText();
		}

		@Override
		public Iterator<Map<String, Object>> iterator() {
			return rows.iterator();
		}

		public List<Map<String,Object>> getJson() {
			return json;
		}

		private List<Map<String, Object>> createJson() {
			final List<Map<String, Object>> rows = new ArrayList<>();
			for (Map<String, Object> row : this) {
				final LinkedHashMap<String, Object> newRow = new LinkedHashMap<>();
				for (String column : columns) {
					final Object value = row.get(column);
					newRow.put(column, toJsonCompatible(value));
				}
				rows.add(newRow);
			}
			return rows;
		}

		private Object toJsonCompatible(Object value) {
			if (value instanceof Node) {
				final Node node = (Node) value;
				final Map<String, Object> result = SubGraph.toMap((PropertyContainer)node);
				result.put("_id",node.getId());

				final List<String> labelNames = SubGraph.getLabelNames(node);
				if (!labelNames.isEmpty()) result.put("_labels", labelNames);
				return result;
			}
			if (value instanceof Relationship) {
				final Relationship relationship = (Relationship) value;
				final Map<String, Object> result = SubGraph.toMap((PropertyContainer) relationship);
				result.put("_id",relationship.getId());
				result.put("_start",relationship.getStartNode().getId());
				result.put("_end",relationship.getEndNode().getId());
				result.put("_type",relationship.getType().name());
				return result;
			}
			if (value instanceof Map) {
				@SuppressWarnings("unchecked") Map<String,Object> map = (Map<String,Object>) value;
				final Map<String,Object> result = new LinkedHashMap<>(map.size());
				for (Map.Entry<String,Object> entry : map.entrySet()) {
					result.put(entry.getKey(), toJsonCompatible(entry.getValue()));
				}
				return result;
			}
			if (value instanceof Iterable) {
				final List<Object> result = new ArrayList<>();
				for (Object inner : (Iterable)value) {
					result.add(toJsonCompatible(inner));
				}
				return result;
			}
			return value;
		}

		public long getTime() {
			return time;
		}

		/*public String getAtsData() {
			return atsData;
		}

		public void setAtsData(String _atsData) {
			atsData=_atsData;
		}*/
	}

	public String prettify(String query) {
		return executionEngine.prettify(query).replaceAll("\n","\n ");
	}
	public CypherResult cypherQuery(String query, String version) {
		// query = replaceIndex(query);
		if (version==null || version.isEmpty() || startsWithCypher(query)) return cypherQuery(query);
		return cypherQuery("CYPHER "+version+" "+query);
	}

	private boolean startsWithCypher(String query) {
		String q = query.trim();
		return q.length() > CYPHER_LENGTH && q.substring(0, CYPHER_LENGTH).equalsIgnoreCase("cypher");
	}

	private CypherResult cypherQuery(String query) {
		if (isMutatingQuery(query)) {
			registerProperties(query);
		}
		boolean canProfile = canProfileQuery(query);
		try {
			return doExecuteQuery(query, canProfile);
		} catch (NotImplementedError |AssertionError e) {
			return doExecuteQuery(query, false);
		}
	}

	private CypherResult doExecuteQuery(String query, boolean canProfile) {
		long time=System.currentTimeMillis();
		org.neo4j.graphdb.Transaction tx = gdb.beginTx();
		javax.transaction.Transaction resumeTx;
		try {
			resumeTx = suspendTx(query);
			ExecutionResult result = canProfile ? executionEngine.profile(query) : executionEngine.execute(query);
			final Collection<Map<String, Object>> data = IteratorUtil.asCollection(result);
			time = System.currentTimeMillis() - time;
			resumeTransaction(resumeTx);
			CypherResult cypherResult = new CypherResult(result.columns(), (List<Map<String, Object>>) data, result.getQueryStatistics(), time, canProfile ? result.executionPlanDescription() : null, prettify(query));
			tx.success();
			return cypherResult;
		} finally {
			// This is not very elegant but we do not close the transaction for the remainder of the session
			//tx.close();
			awaitIndexOnline(query);
		}
	}

	private void awaitIndexOnline(String query) {
		if (!isIndexQuery(query)) return;
		try (org.neo4j.graphdb.Transaction tx = gdb.beginTx()) {
			gdb.schema().awaitIndexesOnline(0, TimeUnit.MILLISECONDS);
			tx.success();
		}
	}

	private javax.transaction.Transaction suspendTx(String query) {
		if (!executionEngine.isPeriodicCommitQuery(query)) return null;
		try {
			return transactionManager.suspend();
		} catch (SystemException e) {
			throw new RuntimeException("Error suspending Transaction",e);
		}
	}
	private void resumeTransaction(javax.transaction.Transaction tx) {
		if (tx == null) return;
		try {
			transactionManager.resume(tx);
		} catch (SystemException | InvalidTransactionException e) {
			throw new RuntimeException("Error resuming Transaction "+tx,e);
		}
	}

	boolean canProfileQuery(String query) {
		//        return false;
		Matcher matcher = CANNOT_PROFILE_PATTERN.matcher(query);
		return !matcher.find();
	}

	private void registerProperties(String query) {
		Set<String> properties = extractProperties(query);
		index.registerProperty(properties);
	}

	String replaceIndex(String query) {
		Matcher matcher = INDEX_PATTERN.matcher(query);
		if (!matcher.find()) return query;
		StringBuffer sb=new StringBuffer();
		do  {
			matcher.appendReplacement(sb,"$1:$1_auto_index(");
		} while (matcher.find());
		matcher.appendTail(sb);
		return sb.toString();
	}

	// TODO should get metadata from the cypher query
	// does not take care of quoted, non-identifier properties
	Set<String> extractProperties(String query) {
		final Matcher matcher = PROPERTY_PATTERN.matcher(query);
		final Set<String> properties = new HashSet<>();
		while (matcher.find()) {
			if (matcher.group(2)!=null) properties.add(matcher.group(2));
			if (matcher.group(3)!=null) properties.add(matcher.group(3));
		}
		return properties;
	}
}
